# -*- Mode: Python; coding: utf-8-dos; -*-

import sys; sys.path.insert(0, u"C:\\Program Files\\Mathematical Systems Inc\\S-Quattro Simulation System V5\\lib3")
from psim import *
from graph import graphsendclient as graphclient
import datetime
import hashlib
graphclient.GraphServer.createServer(uri='http://127.0.0.1:51013')

initialize()

setGlobalSeed(0)

# カスタムコード


# シミュレーションパラメータの定義
class SimulationParam:
    def __init__(self,
                 inputDir = u"input\\default",
                 outputDir = u"output\\default"):
        self.parameter1 = next(constantValue(5.0)) # 
        self.parameter2 = next(constantValue(50.0)) # 

# シミュレーターの定義
class Simulator (SimulatorBase):
    def __init__(self,
                 param = SimulationParam(),
                 inputDir = u"input\\default",
                 outputDir = u"output\\default",
                 showGraph = True,
                 start = lambda self: None,
                 warm = lambda self: 0.0,
                 until = lambda self: 100000.0,
                 odesolver = lambda self: dopri54Solver(dt = 0.1, rtol = 0.001, atol = 1e-06),
                 dataDir = u"data"):
        SimulatorBase.__init__(self, param, inputDir, outputDir, showGraph, start, warm, until, odesolver, dataDir)
        # アイテム の定義
        class Item (FlowItem):
            """部品「生成」で生成され、モデル上を流れるアイテムです。
        
        「属性リスト」は、このアイテムの属性を表します。"""
            def __init__(self, param, **keys):
                FlowItem.__init__(self, param, name = u"アイテム", **keys)
                # パラメータの定義
                for (key, val) in keys.items():
                    setattr(self, key, val)
        
        # 資源の定義
        self.rFacility = SimFacility(self, name = u"ファシリティ", capacity = int(self.param.parameter1), queue = FIFO, monitor = True, releaseIfStopped = False)
        
        # 生成 の定義
        class Generate (Widget):
            """フローアイテムを生成します。
        
        「フローアイテム名」は、この［生成］部品が生成するフローアイテムを表します。
        「初期生成時間」は、最初にフローアイテムを生成するのに要する時間を表します。
        「生成時間」は、フローアイテムを生成するのに要する時間を表します。
        「生成時間系列の生成」は生成時間系列の生成方法を表します。
        「1回の生成数」は、1 回の生成で生成されるフローアイテムの数を表します。生成数が 2 以上の場合はフローアイテムリストとして結合します。フローアイテムリストは分割[Split]で分割可能です。
        「最大生成数」は、この［生成］部品が生成するフローアイテムの最大個数を表します。"""
            def __init__(self, parent):
                Widget.__init__(self, parent, u"生成", nthread = 1)
                # 入力ポートの定義
                # なし
                # 出力ポートの定義
                self.output1 = OutputPort(self, u"出力1", size = 0, save = False, selector = selectRandomUnif(seed = None))
                # パラメータの定義
                self.type = Item # フローアイテム名
                self.firstTime = (lambda dummy: constantValue(0.0)) # 初期生成時間
                self.generationTime = (lambda dummy: exponentialDistribution(mean = 60.0, seed = None)) # 生成時間
                self.regenerateGenerationTimeSeries = 0 # 生成時間系列の生成
                self.eachGenerationNum = (lambda dummy: constantValue(1.0)) # 1回の生成数
                self.maxGenerationNum = None # 最大生成数
            # 動作の定義
            def run(self):
                # 初期設定
                if self.regenerateGenerationTimeSeries:
                    self.generationTime0 = self.generationTime(None)
                    self.generationTime = lambda dummy: self.generationTime0
                
                # 状態の定義
                    
                # 初期生成待ち状態
                def waiting():
                    # 初期生成時間のチェック
                    try:
                        firstTime = float(next(self.firstTime(None)))
                        if firstTime < 0:
                            raise Exception()
                    except ValueError:
                        warn("[%s]部品警告 : 初期生成時間が数値ではありません。初期生成時間を 0 とみなします。" % self.name)
                        firstTime = 0
                    except:
                        warn("[%s]部品警告 : 初期生成時間が 0 以上ではありません。初期生成時間を 0 とみなします。" % self.name)
                        firstTime = 0
                    
                    # 初期生成時間待った後に、生成状態へ
                    yield pause(firstTime) >> \
                        go(generating)(0)
                
                # 生成状態
                def generating(counter): 
                    # 1回の生成数のチェック
                    try:
                        eachGenerationNum = int(next(self.eachGenerationNum(None)))
                        if eachGenerationNum <= 0:
                            raise Exception()
                    except ValueError:
                        fatal("[%s]部品エラー : 1 回の生成数が整数ではありません" % self.name, 
                              self.__class__.__name__ + "'s parameter[eachGenerationNum] is not integer.")
                    except:
                        fatal("[%s]部品エラー : 1 回の生成数が 0 より大きくありません" % self.name,
                              self.__class__.__name__ + "'s parameter[eachGenerationNum] is not greather than 0.")
                
                    # 無限または有限回、アイテムの生成を繰り返す
                    if self.maxGenerationNum is not None and self.maxGenerationNum is not None and self.maxGenerationNum > 0: # 最大生成数が有限であれば
                        if counter < self.maxGenerationNum: # 生成回数が最大生成数よりも小さければ
                            x = min(self.maxGenerationNum - counter, eachGenerationNum)
                            counter += x
                            yield  go(sending)([self.type(self.param) for i in range(x)],counter)
                    elif self.maxGenerationNum is None: # 最大生成数が無限であれば
                        yield go(sending)([self.type(self.param) for i in range(eachGenerationNum)], counter)
                    else:                         
                        warn("[%s]部品警告 : 最大生成数が 0 より大きくありません。最大生成数を無限とみなします。" % self.name)
                        self.maxGenerationNum = None
                        yield go(sending)([self.type(self.param) for i in range(eachGenerationNum)], counter)
                        
                # 送信状態
                def sending(items, counter):
                    # 生成時間のチェック
                    try:
                        generationTime = float(next(self.generationTime(None)))
                        if generationTime < 0:
                            raise Exception()
                    except StopIteration:
                        return
                    except ValueError:
                        fatal("[%s]部品エラー : 生成時間が数値ではありません" % self.name, 
                              self.__class__.__name__ + "'s parameter[generationTime] is not numeric.")
                    except:
                        fatal("[%s]部品エラー : 生成時間が 0 より大きくありません" % self.name,
                              self.__class__.__name__ + "'s parameter[generationTime] is not greather than 0.")
                    
                    # 送るアイテム数が 1 つかどうか
                    if len(items)==1: # 1 つであれば
                        yield self.output1.send(items[0]) >> \
                            pause(generationTime) >> \
                            go(generating)(counter)
                    else:             # でなければ
                        yield self.output1.send(FlowItemList(items)) >> \
                            pause(generationTime) >> \
                            go(generating)(counter)
                
                # 初期生成待ち状態へ遷移
                yield go(waiting)()
        
        # ファシリティ利用 の定義
        class UseFacility (Widget):
            """ファシリティを指定した時間だけ利用します。
        
        「ファシリティリスト」は、利用するファシリティを表します。
        「待ち受け方法」は、ファシリティの待ち受け方法を表します。
        ― \"AND\"は、順番に関わらず全てのファシリティに要求を行ないます。
        ― \"OR\"は、順番に関わらずいずれかのファシリティに要求を行ないます。
        ― \"SEQUENCE\"は、指定したファシリティに指定した順に要求を行ないます。
        「ファシリティ利用時間」は、取得したファシリティを利用する時間を表します。
        「優先度」は、ファシリティを利用する際の優先度を表します。
        「最大待ち時間」は、ファシリティ取得までに待つ最大の待ち時間を表します。
        「並列利用」は、ファシリティの利用を並列的に行うか逐次的に行うかを表します。
        ― \"逐次\"は、ファシリティに対して 1 つだけリクエストを出します。
             リクエストが成立するまで、新しいリクエストは出しません。
        ― \"並列\"は、ファシリティに対して指定した並列数の（複数の）リクエストを出します。
             Infを指定すると並列数の上限なくリクエストを出します。
        「ファシリティ利用の記録」は、ファシリティの利用の可否とファシリティを利用するまでに要した時間の記録の仕方を表します。
        ― \"なし\"は、記録しません。
        ― \"モニター\"は、ファシリティの利用の可否とファシリティの利用に要した時間を
             記録します。
        ― \"時系列モニター\"は、ファシリティの利用の可否とファシリティの利用に要した時間を
             時系列的に記録します。"""
            def __init__(self, parent):
                Widget.__init__(self, parent, u"ファシリティ利用", nthread = 1)
                # 入力ポートの定義
                self.input1 = InputPort(self, u"入力1", size = 10, save = False, selector = selectRandomUnif(seed = None))
                # 出力ポートの定義
                self.output1 = OutputPort(self, u"出力1", size = 0, save = False, selector = selectRandomUnif(seed = None))
                # パラメータの定義
                self.facilityList = [
                    self.parent.rFacility
                ] # ファシリティリスト
                self.waitMethod = 0 # 待ち受け方法
                self.useTime = (lambda item: exponentialDistribution(mean = float(param.parameter2), seed = None)) # ファシリティの利用時間
                self.priority = (lambda item: constantValue(0.0)) # 優先度
                self.maxWait = ['notSpecified', '""', True] # 最大待ち時間
                self.parallel = ['parallel', 'self.param.parameter1', (lambda item: int(self.param.parameter1))] # 並列処理
                self.monitor = 0 # ファシリティ利用の記録
            # 動作の定義
            def run(self):
                # モニターの定義
                if self.monitor == 1:
                    monitor = Monitor(["待ち受け","待ち時間"],["o","f"],name = "%s-%s" %(self.name,"ファシリティの記録"))
                    self.addMonitor(monitor)
                    self.observe = lambda x : monitor.observe(*x)
                elif self.monitor == 2:
                    monitor = TimeMonitor(["待ち受け","待ち時間"],["o","f"],name = "%s-%s" %(self.name,"ファシリティの記録"))
                    self.addMonitor(monitor)
                    self.observe = lambda x : monitor.observe(now(), *x)
                
                if self.maxWait[0] == 'notSpecified':
                    maxWait = lambda item : None
                    outputPort = None
                else:
                    maxWait = self.maxWait[2]['time']
                    outputPort = self.maxWait[2]['outputport']
                
                waitMethods = [allFacilityOf, anyFacilityOf, sequenceFacilityOf]
                
                
                # 状態の定義
                # ファシリティ要求状態
                def requesting(item):
                    t = item.getTime([self.name, self.input1.name, "in"])
                    
                    try:
                        priority = float(next(self.priority(item)))
                    except:
                        warn("[%s]部品警告 : 優先度が数値ではありません。優先度を 0 とみなします。" % self.name)
                        priority = 0
                    
                    if maxWait(item) != None and maxWait(item) < 0:
                        warn("[%s]部品警告 : 最大待ち時間が 0 以上ではありません。最大待ち時間をなしとみなします。" % self.name)
                
                    result = yield waitMethods[self.waitMethod](*self.facilityList).request(name="use", priority=priority) | \
                                    (case(maxWait(item) is not None and maxWait(item) >= 0) >> pause(maxWait(item), name = "time out"))    
                    if "use" in result:
                        if self.monitor > 0:
                            self.observe(["use", now() - t])
                        yield call(using)(result, item)
                        port = self.output1
                    else:
                        if self.monitor > 0:
                            self.observe(["time out", now() - t])
                        port = outputPort if not outputPort is None else self.output1
                    
                    result.releaseAllFacilities()
                    yield go(sending)(port, item)
                
                # ファシリティ利用状態
                def using(result, item):
                    try:
                        useTime = float(next(self.useTime(item)))
                        if useTime < 0:
                            raise Exception()
                    except ValueError:
                        warn("[%s]部品警告 : 利用時間が数値ではありません。利用時間を 0 とみなします。" % self.name)
                        useTime = 0
                    except:
                        warn("[%s]部品警告 : 利用時間が 0 以上ではありません。利用時間を 0 とみなします。" % self.name)
                        useTime = 0
                        
                    yield pause(useTime)
                    
                
                
                # 送信状態
                def sending(port, item):
                    yield port.send(item)
                
                # アイテム受け取り状態
                def waiting():
                    item = (yield self.input1.receive(name = "item"))["item"]
                
                    if self.parallel[1]=="Inf":
                        yield subactivate(requesting)(item) >> \
                            go(waiting)()
                    else:
                        yield call(requesting)(item) >> \
                            go(waiting)()
                    
                # 初期状態へ遷移
                if self.parallel[0]=="parallel":
                    if self.parallel[1]=="Inf":
                        yield go(waiting)()
                    else:
                        for i in range(self.parallel[2](1)):
                            activate(waiting)()
                        yield alwaysTrue()
                else:
                    yield go(waiting)()
        
        # 終端 の定義
        class Terminate (Widget):
            """フローの終端です。
        出力はなく、入力されたフローアイテムは破棄されます。
        
        「待ち時間」は、フローアイテムを破棄してから再びフローアイテムを受け取るまでの最短時間を表します。"""
            def __init__(self, parent):
                Widget.__init__(self, parent, u"終端", nthread = 1)
                # 入力ポートの定義
                self.input1 = InputPort(self, u"入力1", size = 0, save = False, selector = selectRandomUnif(seed = None))
                # 出力ポートの定義
                # なし
                # パラメータの定義
                self.waitTime = constantValue(0.0) # 待ち時間
            # 動作の定義
            def run(self):
                # 状態の定義
                
                # アイテム受け取り待ち状態
                def waiting():
                   # パラメータチェック
                    try:
                        waitTime = float(next(self.waitTime))
                        if waitTime < 0:
                            raise Exception()
                    except ValueError:
                        warn("[%s]部品警告 : 待ち時間が数値ではありません。待ち時間を 0 とみなします。" % self.name)
                        waitTime = 0
                    except:
                        warn("[%s]部品警告 : 待ち時間が 0 以上ではありません。待ち時間を 0 とみなします。" % self.name)
                        waitTime = 0
                
                    yield pause(waitTime) >> self.input1.receive(name = "item") >> go(waiting)()
                    
                # 初期状態へ遷移
                yield go(waiting)()
        
        # オブジェクトの生成
        self.wGenerate = Generate(self)
        self.wUseFacility = UseFacility(self)
        self.wTerminate = Terminate(self)
        # リンクの生成
        Link(self.wGenerate.output1, self.wUseFacility.input1)
        Link(self.wUseFacility.output1, self.wTerminate.input1)
    def run(self, **args):
        # オブジェクトの起動
        try:
            _sim_before(self)
        except:
            pass
        activate(self.wGenerate.start)()
        activate(self.wUseFacility.start)()
        activate(self.wTerminate.start)()
        def startProcs():
            for proc in self.procs:
                proc()
            yield alwaysTrue()
        activate(startProcs)()
        # シミュレーション開始
        start(until = self.until, odesolver = self.odesolver, **args)
        # 結果の保存
        self.save()
        self.flush()
        try:
            _sim_after(self)
        except:
            pass
    def objective(self):
        return TimeMonitor(name = u"ファシリティ", basedir = self.outputDir)[u"バッファ"].mean()

if __name__ == "__main__":
    # 実行時の処理
    simulator = Simulator(param = SimulationParam(),
        inputDir = u"input\\default",
        outputDir = u"output\\default",
        start = lambda self: None,
        warm = lambda self: 0.0,
        until = lambda self: 100000.0,
        odesolver = lambda self: dopri54Solver(dt = 0.1, rtol = 0.001, atol = 1e-06))
    
    simulator.run()
    result = simulator.calcObjective()
    print(result)
    simulator.saveResult(result, u"objective")
    if simulator.agentApp is not None:
        print(u"エージェント画面の終了待ち受け...")
    simulator.join()
    sys.exit()
